﻿// FindSlice.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#include <opencv2/core.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/features2d.hpp>

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/calib3d/calib3d.hpp>

#if _DEBUG
#pragma comment(lib, "opencv_world320d.lib")
#else
#pragma comment(lib, "opencv_world320.lib")
#endif

// libtiff
#include "tiff\tiffio.h"
#pragma comment(lib,"libtiff.lib")

#include <map>
#include <list>
#include <mutex>
#include <numeric>
#include <iostream>
#include <fstream>
#include <Windows.h>


cv::Mat NormalizeHist(cv::Mat data, double gamma = 1.0)
{
	double max_val, min_val;
	cv::minMaxLoc(data, &min_val, &max_val);

	int image_num = 1;
	int channels[] = { 0 };
	cv::MatND hist; 
	int dim_num = 1;
	int bin_num = ceil(max_val - min_val);
	int bin_nums[] = { bin_num };
	float range[] = { min_val, max_val };
	const float *ranges[] = { range };

	cv::calcHist(&data, image_num, channels, cv::Mat(), hist, dim_num, bin_nums, ranges);

	int scale_max = 0;
	int scale_min = 0;
	int scale_peak1 = 0;
	int scale_peak2 = 0;
	std::list<int> mark;
	const float max_th = data.cols * data.rows / 15;
	float sum = 0;
	for (int i = bin_num - 1; i >= 0; i--) {
		scale_max = i;
		if ((sum += hist.at<float>(i)) > max_th)
			break;
	}

	float max1 = 0, min1, max2;
	bool init = true;
	for (int i = 0; i < bin_num; i++) {
		auto value = hist.at<float>(i);
		if (value > max1) {
			init = true;
			max1 = value;
			min1 = value;
			max2 = 0;
			scale_peak1 = i;
		}
		else {
			if (init && value <= min1) {
				min1 = value;
				max2 = value;
				scale_min = i;
				init = true;
			}
			else {
				if (min1 < max1 / 4 && value > min1 + MIN((max1 - min1) / 100, 100)) {
					init = false;
					if (value > max2) {
						max2 = value;
						scale_peak2 = i;
					}
				}
			}
		}
	}
	for (int i = scale_peak2; i >= scale_peak1; i--) {
		auto value = hist.at<float>(i);
		if (value < min1) {
			min1 = value;
			scale_min = i;
		}
	}

	if (scale_peak2 >= scale_max) {
		scale_peak2 = scale_max - 1;
	}

	if (scale_peak1 < scale_min &&
		scale_min < scale_peak2 &&
		scale_peak2 < scale_max) {
		//ok
		//printf("OK %d %d %d %d\n", scale_peak1, scale_min, scale_peak2, scale_max);
	}
	else {
		printf("ERROR %d %d %d %d\n", scale_peak1, scale_min, scale_peak2, scale_max);
		return data;
	}
	int a1 = (scale_peak2 - scale_peak1) * 2 / 5 + scale_peak1;
	if (a1 < scale_min) scale_min = a1;
	scale_max = scale_max * (max_val - min_val) / bin_num + min_val;
	scale_min = scale_min * (max_val - min_val) / bin_num + min_val;
	scale_peak1 = scale_peak1 * (max_val - min_val) / bin_num + min_val;
	scale_peak2 = scale_peak2 * (max_val - min_val) / bin_num + min_val;

	cv::Mat result;
	double alpha1 = 100 / (scale_peak2 - scale_min);
	double alpha2 = 200 / (scale_max - scale_min);
	double alpha = (alpha1 + alpha2) / 2;
	if (!isfinite(alpha)) alpha = 1;
	((cv::Mat)(data - scale_min)).convertTo(result, CV_8U, alpha);

	uchar lut[256];
	double gm = 1.0 / gamma;
	for (int i = 0; i < 256; i++)
	{
		lut[i] = pow(1.0*i / 255, gm) * 255;
	}
	cv::Mat result2;
	cv::LUT(result, cv::Mat(cv::Size(256, 1), CV_8U, lut), result2);

	return result2;
}


std::string fixpath(std::string path)
{
	auto i1 = path.find_last_of('\\');
	auto i2 = path.find_last_of('/');
	if (i1 == std::string::npos) {
		if (i2 == std::string::npos)
			return path;
		else if (i2 < path.length() - 1)
			return path + "/";
		else
			return path;
	}
	else {
		if (i1 < path.length() - 1)
			return path + "\\";
		else
			return path;
	}
}

// sort by number
std::vector<std::string> GetFileList_num(std::string path) 
{
	path = fixpath(path);
	std::multimap<int, std::string> filelist;
	{
		WIN32_FIND_DATAA fd;
		HANDLE hdir = FindFirstFileA((path + "*.tif*").c_str(), &fd);
		if (hdir != INVALID_HANDLE_VALUE) {
			do {
				std::string filename(fd.cFileName);
				auto index = filename.find_first_of("0123456789");
				int num = (index == std::string::npos) ? 0 : std::stoi(filename.substr(index));
				filelist.emplace(num, path + filename);
			} while (FindNextFileA(hdir, &fd));
			FindClose(hdir);
		}
	}
	std::vector<std::string> files;
	for (auto it = filelist.begin(); it != filelist.end(); ++it)
	{
		files.push_back(it->second);
	}
	return files;
}

// list as is
std::vector<std::string> GetFileList(std::string path)
{
	path = fixpath(path);
	std::vector<std::string> files;
	{
		WIN32_FIND_DATAA fd;
		HANDLE hdir = FindFirstFileA((path + "*.tif*").c_str(), &fd);
		if (hdir != INVALID_HANDLE_VALUE) {
			do {
				files.push_back(path + fd.cFileName);
			} while (FindNextFileA(hdir, &fd));
			FindClose(hdir);
		}
	}
	return files;
}

cv::Mat LoadMultiTiff(std::string filename) 
{
	cv::Mat result;
	TIFFErrorHandler oldhandler;
	oldhandler = TIFFSetWarningHandler(NULL);

	int StackCount = 0;
	WORD *outpixel = NULL;
	int width = 0;
	int height = 0;
	TIFF *in = TIFFOpen(filename.c_str(), "r");
	if (!in) {
		printf("error on open tiff\n");
		return result;
	}
	if (StackCount == 0) {
		do {
			StackCount++;
		} while (TIFFReadDirectory(in));
		TIFFSetDirectory(in, 0);
	}

	int stack = 0;
	do {
		int nImageLength = 0;
		int nImageWidth = 0;
		int nBitsPerSample = 1;
		int nSamplePerPixel = 1;

		if (!TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &nImageWidth) ||
			!TIFFGetField(in, TIFFTAG_IMAGELENGTH, &nImageLength) ||
			!TIFFGetField(in, TIFFTAG_BITSPERSAMPLE, &nBitsPerSample) ||
			!TIFFGetField(in, TIFFTAG_SAMPLESPERPIXEL, &nSamplePerPixel))
		{
			printf("error in tiff tag\n");
			return result;
		}

		if (nBitsPerSample != 16) {
			printf("unsupported BitsPerSample %d\n", nBitsPerSample);
			return result;
		}
		if (nSamplePerPixel != 1) {
			printf("unsupported SamplePerPixel %d\n", nSamplePerPixel);
			return result;
		}

		if (!outpixel) {
			width = nImageWidth;
			height = nImageLength;
			int64 outsize = (int64)width*height;
			try {
				outpixel = new WORD[outsize]();
			}
			catch (std::bad_alloc) {
				fprintf(stderr, "memory allocation error.\n");
				return result;
			}
		}

		WORD *buf = (WORD *)_TIFFmalloc(TIFFScanlineSize(in));
		if (!buf) {
			fprintf(stderr, "memory allocation error.\n");
			return result;
		}

		for (int y = 0; y < nImageLength; y++) {
			TIFFReadScanline(in, buf, y);
			for (int x = 0; x < nImageWidth; x++) {
				WORD value1 = buf[x];
				int64 index = x + y * width;//x * height + y;																				  
				WORD value2 = outpixel[index];
				outpixel[index] = max(value1, value2);
			}
		}
		_TIFFfree(buf);
		stack++;
	} while (TIFFReadDirectory(in));
	TIFFClose(in);

	TIFFSetWarningHandler(oldhandler);

	result = cv::Mat(height, width, CV_16U);
	auto p_dst = result.ptr<WORD>(0);
	auto p_src = outpixel;
	for (int64 i = 0; i < width*height; i++) {
		*p_dst++ = *p_src++;
	}
	delete[] outpixel;
	return result;
}

std::string GetFilenameFromPath(std::string path)
{
	int ind = path.find_last_of("\\/");
	return ((ind == std::string::npos) ? path : path.substr(ind + 1));
}

int main(int argc, const char* argv[])
{
	if (argc < 3) {
		printf("%s [db input dir] [query input dir] (gamma=1.0) (downsize=16.0) (dilate=2)\n", argv[0]);
		return 0;
	}
	std::string db_path(argv[1]);
	std::string query_path(argv[2]);

	// reduction rate
	double downsize = 16.0;
	// maximum filter
	int dilate_kernel = 2;
	// gamma for brightness control
	double gamma = 1.0;

	if (argc > 3) {
		std::stringstream str(argv[3]);
		str >> gamma;
	}
	if (argc > 4) {
		std::stringstream str(argv[4]);
		str >> downsize;
	}
	if (argc > 5) {
		std::stringstream str(argv[5]);
		str >> dilate_kernel;
	}
	printf("gamma:%f\n", gamma);
	printf("down scale factor:%f\n", downsize);
	printf("dilate kernel size:%d\n", dilate_kernel);

	// directory for whole brain
	auto dbfiles = GetFileList_num(db_path);
	// directory for query files
	auto queryfiles = GetFileList_num(query_path);

	int x_size = 0, y_size = 0;
	int x_count = 0, y_count = 0;
	int count = dbfiles.size();
	IplImage* iplImage = NULL;

	printf("database files loading...\n");
	int i = 0;
	for each(auto dbfile in dbfiles) {
		printf("process %s\n", dbfile.c_str());
		{
			cv::Mat downsample;
			cv::Mat normalized;
			{
				// load tiff stack as max intensity
				cv::Mat orgfile = LoadMultiTiff(dbfile);
				cv::Mat filtered;
				cv::dilate(orgfile, filtered, cv::Mat::ones(dilate_kernel, dilate_kernel, CV_8UC1));
				cv::resize(filtered, downsample, cv::Size(), 1 / downsize, 1 / downsize, cv::InterpolationFlags::INTER_AREA);
				normalized = NormalizeHist(downsample, gamma);
			}
			if (x_size == 0 && y_size == 0) {
				x_size = normalized.cols;
				y_size = normalized.rows;

				if (count < 3) {
					y_count = 1;
					x_count = count;
				}
				else {
					x_count = (int)(sqrt(count)) + 1;
					y_count = count / x_count + 1;
					if (x_count * y_count - count > 2) {
						int sp = x_count * y_count - count;
						for (int x = x_count - 10; x < x_count + 10; x++) {
							if (x < 1) continue;
							int y = count / x + 1;
							if (y < 2) break;
							if (x * y - count < sp) {
								x_count = x;
								y_count = y;
								sp = x_count * y_count - count;
							}
						}
					}
				}

				iplImage = cvCreateImage(cvSize(x_size * x_count, y_size * y_count), IPL_DEPTH_8U, 1);
				cvSet(iplImage, CvScalar(0));
			}

			int y = i / x_count;
			int x = i - y * x_count;
			{
				IplImage aImg = normalized;
				cvSetImageROI(iplImage, cvRect(x * x_size, y * y_size, aImg.width, aImg.height));
				cvCopy(&aImg, iplImage);
			}
			cvResetImageROI(iplImage);
		}
		i++;
	}

	cv::Mat dbImage(cv::cvarrToMat(iplImage));

	printf("target files loading...\n");
	for each(auto qfile in queryfiles) {
		printf("process %s\n", qfile.c_str());
		{
			cv::Mat QImg;
			{
				cv::Mat orgfile = LoadMultiTiff(qfile);
				cv::Mat filtered;
				cv::dilate(orgfile, filtered, cv::Mat::ones(dilate_kernel, dilate_kernel, CV_8UC1));
				cv::Mat downscale;
				cv::resize(filtered, downscale, cv::Size(), 1 / downsize, 1 / downsize, cv::InterpolationFlags::INTER_AREA);
				QImg = NormalizeHist(downscale, gamma);
			}
			cv::Mat result;
			{
				cv::Mat target, templ;
				target = dbImage;
				templ = QImg;
				cv::matchTemplate(target, templ, result, cv::TM_CCORR_NORMED);
			}
			cv::Mat result_disp;
			cv::cvtColor(dbImage, result_disp, CV_GRAY2RGB);

			const int listn = 5;
			for (int i = 0; i < listn; i++) {
				double corr_max;
				cv::Point max_pt;
				cv::minMaxLoc(result, NULL, &corr_max, NULL, &max_pt);
				int x_detect = (int)((double)max_pt.x / x_size + 0.5);
				int y_detect = (int)((double)max_pt.y / y_size + 0.5);
				int detect_idx = y_detect * x_count + x_detect;
				std::string foundFilename(GetFilenameFromPath(dbfiles[detect_idx]));
				printf("no.%d corr=%f (x:%d(%d), y:%d(%d)):%s\n", i+1, corr_max, max_pt.x, x_detect, max_pt.y, y_detect, foundFilename.c_str());

				cv::Rect roi_detect(max_pt.x, max_pt.y, QImg.cols, QImg.rows);
				cv::Scalar color(0, 255 * (listn - i) / listn, 255 * (listn - i) / listn);
				cv::rectangle(result_disp, roi_detect, color, 3 + (listn - i));
				{
					std::stringstream str;
					str << i + 1 << " : " << corr_max;
					cv::Point text_pt(max_pt);
					text_pt.y += QImg.rows - 5;
					text_pt.x += 5;
					cv::putText(result_disp, str.str(), text_pt, cv::FONT_HERSHEY_PLAIN, 5, color, 8);
				}
				{
					cv::Point text_pt(max_pt);
					int baseline = 0;
					cv::Size str_size = cv::getTextSize(foundFilename, cv::FONT_HERSHEY_PLAIN, 3, 3, &baseline);
					text_pt.y += str_size.height + 5;
					text_pt.x += 5;
					cv::putText(result_disp, foundFilename, text_pt, cv::FONT_HERSHEY_PLAIN, 3, cv::Scalar(255,255,255), 5);
				}
				// remove found peek and search next peek
				for (int x = MAX(max_pt.x - QImg.cols / 2, 0); x < MIN(max_pt.x + QImg.cols / 2, result.cols); x++) {
					for (int y = MAX(max_pt.y - QImg.rows / 2, 0); y < MIN(max_pt.y + QImg.rows / 2, result.rows); y++) {
						result.at<float>(y, x) = 0;
					}
				}
			}

			std::string outfile = "result_" + GetFilenameFromPath(qfile);
			printf("write result: %s\n", outfile.c_str());
			cv::imwrite(outfile.c_str(), result_disp);
		}
	}
	cvReleaseImage(&iplImage);

	return 0;
}
